
import time
import os
import logging
from os.path import isdir, join, relpath, pardir, exists
from mimetypes import guess_type

from .exceptions import HttpNotFound, HttpMethodNotAllowed, HttpForbidden
from .contexts import context
from .constants import HTTP_DATETIME_FORMAT


logging.basicConfig(level=logging.INFO)


class Controller(object):
    __http_methods__ = 'any'
    __response_encoding__ = 'utf8'
    __default_action__ = 'index'
    __remove_trailing_slash__ = True

    # noinspection PyMethodMayBeStatic
    def _serve_handler(self, handler, *remaining_paths):
        if hasattr(handler, '__response_encoding__'):
            context.response_encoding = handler.__response_encoding__

        if hasattr(handler, '__content_type__'):
            context.response_content_type = handler.__content_type__

        return handler(*remaining_paths)

    def _dispatch(self, *remaining_paths):
        if not len(remaining_paths):
            path = self.__default_action__
        else:
            path = self.__default_action__ if remaining_paths[0] == '' else remaining_paths[0]
            remaining_paths = remaining_paths[1:]

        # Ensuring the handler
        handler = getattr(self, path, None)
        if handler is None:
            handler = getattr(self, self.__default_action__, None)
            if handler is not None:
                remaining_paths = (path, ) + remaining_paths

        args_count = len(remaining_paths)
        if (handler is None or not hasattr(handler, '__http_methods__')) \
                or (hasattr(handler, '__argcount__') and handler.__argcount__ < args_count) \
                or (hasattr(handler, '__annotations__') and len(handler.__annotations__) < args_count):
            raise HttpNotFound()

        if 'any' != handler.__http_methods__ and context.method not in handler.__http_methods__:
            raise HttpMethodNotAllowed()

        return handler, remaining_paths

    def __call__(self, *remaining_paths):
        handler, remaining_paths = self._dispatch(*remaining_paths)
        return self._serve_handler(handler, *remaining_paths)


class RestController(Controller):
    __detect_verb_by_header__ = False

    def _dispatch(self, *remaining_paths):
        handler = None
        if remaining_paths:
            first_path = remaining_paths[0]
            if hasattr(self, first_path):
                remaining_paths = remaining_paths[1:]
                handler = getattr(self, first_path)

        if handler is None:

            if self.__detect_verb_by_header__ and context.environ.get(self.__detect_verb_by_header__):
                handler = getattr(self, context.environ.get(self.__detect_verb_by_header__).lower())
            elif not hasattr(self, context.method):
                raise HttpMethodNotAllowed()
            else:
                handler = getattr(self, context.method)

        if handler is None or not hasattr(handler, '__http_methods__'):
            raise HttpNotFound()

        # FIXME: check argcount
        if hasattr(handler, '__annotations__') and len(handler.__annotations__) < len(remaining_paths):
            raise HttpNotFound()

        return handler, remaining_paths


class Static(Controller):
    __response_encoding__ = None
    __chunk_size__ = 0x4000

    def __init__(self, directory='.', default_document='index.html'):
        self.default_document = default_document
        self.directory = directory

    def __call__(self, *remaining_paths):

        # Find the physical path of the given path parts
        physical_path = join(self.directory, *remaining_paths)

        # Check to do not access the parent directory of root and also we are not listing directories here.
        if pardir in relpath(physical_path, self.directory):
            raise HttpForbidden()

        if isdir(physical_path):
            if self.default_document:
                physical_path = join(physical_path, self.default_document)
                if not exists(physical_path):
                    raise HttpForbidden
            else:
                raise HttpForbidden()

        context.response_headers.add_header('Content-Type', guess_type(physical_path)[0] or 'application/octet-stream')

        try:
            f = open(physical_path, mode='rb')
            stat = os.fstat(f.fileno())
            context.response_headers.add_header('Content-Length', str(stat[6]))
            context.response_headers.add_header(
                'Last-Modified',
                time.strftime(HTTP_DATETIME_FORMAT, time.gmtime(stat.st_mtime))
            )

            with f:
                while True:
                    r = f.read(self.__chunk_size__)
                    if not r:
                        break
                    yield r

        except OSError:
            raise HttpNotFound()




import unittest

from nanohttp import Controller, html, json, RestController, text
from nanohttp.tests.helpers import WsgiAppTestCase


class LinksController(Controller):

    @html
    def index(self, article_id: int=None, link_id: int=None):
        if link_id:
            yield 'Article: %s, Links index: %s' % (article_id, link_id)
        else:
            yield 'Link1, Link2, Link3'


class ArticleController(RestController):
    __detect_verb_by_header__ = 'HTTP_X_HTTP_VERB'
    links = LinksController()

    @html
    def get(self, article_id: int=None):
        yield "GET Article%s" % (
            's' if not article_id else (': ' + article_id)
        )

    @html
    def post(self):
        yield "POST Article"

    @html
    def put(self, article_id: int=None, inner_resource: str=None, link_id: int=None):
        if inner_resource == 'links':
            yield from self.links(article_id, link_id)
        else:
            yield "PUT Article: %s" % article_id

    def disallowed(self):  # pragma: no cover
        yield 'bad'


class DispatcherTestCase(WsgiAppTestCase):

    class Root(Controller):

        articles = ArticleController()

        @html
        def index(self):
            yield 'Index'

        @json(verbs='post')
        def login(self):
            return ["token"]

        @html(verbs=['get', 'post'])
        def contact(self, contact_id: int=None):
            yield "Contact: %s" % contact_id

        @html
        def users(self, user_id: int=None, attr: str=None):
            yield 'User: %s\n' % user_id
            yield 'Attr: %s\n' % attr

        @html
        def bad(self):
            raise Exception()

        @text
        def empty(self):
            pass

    def test_root(self):
        self.assert_get('/', expected_response='Index')
        self.assert_get('/index', expected_response='Index')

    def test_arguments(self):
        self.assert_get('/contact/1', expected_response='Contact: 1')
        self.assert_post('/contact', expected_response='Contact: None')
        self.assert_get('/users', expected_response='User: None\nAttr: None\n')
        self.assert_get('/users/10/')
        self.assert_get('/users/10/jobs', expected_response='User: 10\nAttr: jobs\n')
        self.assert_get('/users/10/11/11', status=404)
        self.assert_post('/articles/2', status=404)

    def test_verbs(self):
        self.assert_put('/', expected_response='Index')
        self.assert_post('/login', expected_response='[\n    "token"\n]')
        self.assert_get('/contact/test', expected_response='Contact: test')
        self.assert_post('/contact', expected_response='Contact: None')
        self.assert_put('/contact', status=405)
        self.assert_options('/contact', status=405)
        self.assert_options('/')
        self.assert_delete('/articles', status=405)

    def test_rest_controller(self):
        self.assert_get('/articles', expected_response='GET Articles')
        self.assert_get('/articles/23', expected_response='GET Article: 23')
        self.assert_post('/articles', expected_response='POST Article')
        self.assert_put('/articles/23', expected_response='PUT Article: 23')
        self.assert_put('/articles/23/links/2', expected_response='Article: 23, Links index: 2')
        self.assert_get('/articles/links', expected_response='Link1, Link2, Link3')
        self.assert_request('/articles', 'disallowed', status=404)
        self.assert_get('/articles/disallowed', status=404)

    def test_trailing_slash(self):
        self.assert_get('/users/10/jobs/', expected_response='User: 10\nAttr: jobs\n')

    def test_exception(self):
        self.assert_get('/bad', status=500)

    def test_empty_response(self):
        self.assert_get('/empty', expected_response='')

    def test_detect_verb_by_header(self):
        self.assert_post(
            '/articles/23',
            headers={
                'X-HTTP-VERB': 'put'
            },
            expected_response='PUT Article: 23'
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import re
import ujson
import unittest
from os.path import join

from nanohttp import Controller, text, context
from nanohttp.tests.helpers import WsgiAppTestCase, STUFF_DIR


class FormTestCase(WsgiAppTestCase):
    class Root(Controller):

        @text('post')
        def index(self):
            if not context.form:
                yield 'empty'
            yield ', '.join('%s: %s' % (k, v) for k, v in sorted(context.form.items(), key=lambda x: x[0]))

    def test_simple_query_string(self):
        self.assert_post('/?a=1&b=&c=2', expected_response="a: 1, b: , c: 2")
        self.assert_post('/?a=1&b=2&b=3', expected_response="a: 1, b: ['2', '3']")

    def test_url_encoded(self):
        self.assert_post(
            '/',
            fields={
                'a': 1,
                'b': [2, 3],
            },
            expected_response='a: 1, b: [2, 3]'
        )

    def test_multipart(self):
        self.assert_post(
            '/',
            fields={
                'a': 1,
                'b': [2, 3],
            },
            files={'c': join(STUFF_DIR, 'cat.jpg')},
            expected_response=re.compile("a: 1, b: \[2, 3\], c: FieldStorage\('c', 'cat\.jpg.*"),
        )

    def test_empty_form(self):
        self.assert_post(
            '/',
            fields={},
            expected_response='empty',
        )

    def test_json(self):
        self.assert_post(
            '/',
            json=ujson.dumps({
                'a': 1,
                'b': [2, 3],
            }),
            expected_response='a: 1, b: [2, 3]',
        )

    def test_invalid_form(self):
        self.assert_post(
            '/',
            fields={
                'a': 1,
                'b': [2, 3],
            },
            status=400,
            headers={'Content-Type': 'invalid/content-type'}
        )


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import time
import unittest

from nanohttp import Controller, quickstart, settings
from nanohttp.tests.helpers import WsgiAppTestCase, find_free_tcp_port


class QuickstartTestCase(WsgiAppTestCase):

    class Root(Controller):
        pass

    def setUp(self):
        super().setUp()
        self.port = find_free_tcp_port()

    def test_without_controller(self):
        shutdown = quickstart(port=self.port, block=False)
        self.assertTrue(callable(shutdown))
        time.sleep(.5)
        shutdown()

    def test_with_controller(self):
        shutdown = quickstart(controller=self.Root(), block=False, port=self.port)
        self.assertTrue(callable(shutdown))
        time.sleep(.5)
        shutdown()

    def test_with_application(self):
        shutdown = quickstart(application=self.application, block=False, port=self.port)
        self.assertTrue(callable(shutdown))
        time.sleep(.5)
        shutdown()

    def test_before_configure(self):
        settings.__class__._set_instance(None)
        shutdown = quickstart(
            port=self.port,
            block=False,
            config='''
            test_config_item: item1
            '''
        )
        self.assertTrue(callable(shutdown))
        time.sleep(.5)
        shutdown()


if __name__ == '__main__':  # pragma: no cover
    unittest.main()






from os.path import join

from nanohttp import Controller, Static
from nanohttp.tests.helpers import WsgiAppTestCase, STUFF_DIR, md5sum

CAT = join(STUFF_DIR, 'cat.jpg')


class StaticFileTestCase(WsgiAppTestCase):

    class Root(Controller):
        cat = Static(CAT)
        static = Static(STUFF_DIR)
        assets = Static(STUFF_DIR, default_document=None)

    def test_static_file(self):
        checksum = md5sum(CAT)
        self.assert_get('/cat', expected_checksum=checksum)
        self.assert_get('/static/cat.jpg', expected_checksum=checksum)
        self.assert_get('/static', status=403)
        self.assert_get('/static/not-exists', status=404)
        self.assert_get('/static/..', status=403)
        self.assert_get('/assets/', status=403)
